@@ -43,6 +43,8 @@ from .. import types as pg_types
 from ..types import io as pg_types_io
 from ..types.io import lib as io_lib
 
+import warnings
+
 # Map element3.Notice field identifiers
 # to names used by message.Message.
 notice_field_to_name = {
@@ -148,6 +150,21 @@ class TypeIO(pg_api.TypeIO):
 	def lookup_composite_type_info(self, typid):
 		return self.database.sys.lookup_composite(typid)
 
+	def lookup_domain_basetype(self, typid):
+		if self.database.version_info[:2] >= (8, 4):
+			return self.lookup_domain_basetype_84(typid)
+
+		while typid:
+			r = self.database.sys.lookup_basetype(typid)
+			if not r[0][0]:
+				return typid
+			else:
+				typid = r[0][0]
+
+	def lookup_domain_basetype_84(self, typid):
+		r = self.database.sys.lookup_basetype_recursive(typid)
+		return r[0][0]
+
 	def set_encoding(self, value):
 		"""
 		Set a new client encoding.
@@ -190,7 +207,11 @@ class TypeIO(pg_api.TypeIO):
 		if oid in self.typinfo:
 			nsp, name, *_ = self.typinfo[oid]
 			return qi(nsp) + '.' + qi(name)
-		return 'pg_catalog.' + pg_types.oid_to_name.get(oid)
+		name = pg_types.oid_to_name.get(oid)
+		if name:
+			return 'pg_catalog.%s' % name
+		else:
+			return None
 
 	def type_from_oid(self, oid):
 		if oid in self._cache:
@@ -225,6 +246,8 @@ class TypeIO(pg_api.TypeIO):
 		else:
 			typio = builtins(typid)
 			if typio is not None:
+				# If typio is a tuple, it's a constant pair: (pack, unpack)
+				# otherwise, it's an I/O pair constructor.
 				if typio.__class__ is not tuple:
 					typio = typio(typid, self)
 				self._cache[typid] = typio
@@ -254,16 +277,22 @@ class TypeIO(pg_api.TypeIO):
 					for x in self.lookup_composite_type_info(typrelid):
 						attmap[x[1]] = i
 						attnames.append(x[1])
+						if x[2]:
+							# This is a domain
+							fieldtypid = self.lookup_domain_basetype(x[0])
+						else:
+							fieldtypid = x[0]
 						typids.append(x[0])
-						pack, unpack, typ = self.resolve(
-							x[0], list(from_resolution_of) + [typid]
+						te = self.resolve(
+							fieldtypid, list(from_resolution_of) + [typid]
 						)
-						cio.append((pack or self.encode, unpack or self.decode))
+						cio.append((te[0] or self.encode, te[1] or self.decode))
 						i += 1
 					self._cache[typid] = typio = self.record_io_factory(
 						cio, typids, attmap, list(
 							map(self.sql_type_from_oid, typids)
 						), attnames,
+						typrelid,
 						quote_ident(typnamespace) + '.' + \
 						quote_ident(typname),
 					)
@@ -282,7 +311,21 @@ class TypeIO(pg_api.TypeIO):
 					)
 					self._cache[typid] = typio
 				else:
-					self._cache[typid] = typio = self.strio
+					typio = None
+					if typtype == b'd':
+						basetype = self.lookup_domain_basetype(typid)
+						typio = self.resolve(
+							basetype,
+							from_resolution_of = list(from_resolution_of) + [typid]
+						)
+					elif typtype == b'p' and typnamespace == 'pg_catalog' and typname == 'record':
+						# anonymous record type
+						typio = self.anon_record_io_factory()
+
+					if not typio:
+						typio = self.strio
+
+					self._cache[typid] = typio
 			else:
 				# Throw warning about type without entry in pg_type?
 				typio = self.strio
@@ -349,14 +392,14 @@ class TypeIO(pg_api.TypeIO):
 		if hasbin_output:
 			def unpack_an_array(data, array_from_parts = self.array_from_parts):
 				flags, typoid, dims, lbs, elements = array_unpack(data)
-				return array_from_parts((map(unpack_element, elements), dims, lbs))
+				return array_from_parts(((x if x is None else unpack_element(x) for x in elements), dims, lbs))
 		else:
 			# signals string formatting
 			unpack_an_array = None
 
 		return (pack_an_array, unpack_an_array, pg_types.Array)
 
-	def RowTypeFactory(self, attribute_map = {}, _Row = pg_types.Row.from_sequence):
+	def RowTypeFactory(self, attribute_map = {}, _Row = pg_types.Row.from_sequence, composite_relid = None):
 		return partial(_Row, attribute_map)
 
 	##
@@ -368,6 +411,7 @@ class TypeIO(pg_api.TypeIO):
 		attmap : "mapping of column name to index number",
 		typnames : "sequence of sql type names in order",
 		attnames : "sequence of attribute names in order",
+		composite_relid : "oid of the composite relation",
 		composite_name : "the name of the composite type",
 		get0 = get0,
 		get1 = get1,
@@ -375,9 +419,9 @@ class TypeIO(pg_api.TypeIO):
 	):
 		fpack = tuple(map(get0, column_io))
 		funpack = tuple(map(get1, column_io))
-		row_constructor = self.RowTypeFactory(attribute_map = attmap)
+		row_constructor = self.RowTypeFactory(attribute_map = attmap, composite_relid = composite_relid)
 
-		def raise_pack_tuple_error(procs, tup, itemnum):
+		def raise_pack_tuple_error(cause, procs, tup, itemnum):
 			data = repr(tup[itemnum])
 			if len(data) > 80:
 				# Be sure not to fill screen with noise.
@@ -388,9 +432,9 @@ class TypeIO(pg_api.TypeIO):
 				(b'M', fmt_errmsg('pack', itemnum, attnames[itemnum], typnames[itemnum], composite_name),),
 				(b'W', data,),
 				(b'P', str(itemnum),)
-			)))
+			)), cause = cause)
 
-		def raise_unpack_tuple_error(procs, tup, itemnum):
+		def raise_unpack_tuple_error(cause, procs, tup, itemnum):
 			data = repr(tup[itemnum])
 			if len(data) > 80:
 				# Be sure not to fill screen with noise.
@@ -401,7 +445,7 @@ class TypeIO(pg_api.TypeIO):
 				(b'M', fmt_errmsg('unpack', itemnum, attnames[itemnum], typnames[itemnum], composite_name),),
 				(b'W', data,),
 				(b'P', str(itemnum),),
-			)))
+			)), cause = cause)
 
 		def unpack_a_record(data,
 			unpack = io_lib.record_unpack,
@@ -426,6 +470,32 @@ class TypeIO(pg_api.TypeIO):
 			)
 		return (pack_a_record, unpack_a_record, tuple)
 
+	def anon_record_io_factory(self):
+		def raise_unpack_tuple_error(cause, procs, tup, itemnum):
+			data = repr(tup[itemnum])
+			if len(data) > 80:
+				# Be sure not to fill screen with noise.
+				data = data[:75] + ' ...'
+			self.raise_client_error(element.ClientError((
+				(b'C', '--cIO',),
+				(b'S', 'ERROR',),
+				(b'M', 'Could not unpack element {} from anonymous record'.format(itemnum)),
+				(b'W', data,),
+				(b'P', str(itemnum),)
+			)), cause = cause)
+
+		def _unpack_record(data, unpack = io_lib.record_unpack, process_tuple = process_tuple):
+			record = list(unpack(data))
+			coloids = tuple(x[0] for x in record)
+			colio = map(self.resolve, coloids)
+			column_unpack = tuple(c[1] or self.decode for c in colio)
+
+			data = tuple(x[1] for x in record)
+
+			return process_tuple(column_unpack, data, raise_unpack_tuple_error)
+
+		return (None, _unpack_record)
+
 	def raise_client_error(self, error_message, cause = None, creator = None):
 		m = {
 			notice_field_to_name[k] : v
@@ -488,7 +558,7 @@ class TypeIO(pg_api.TypeIO):
 				# This should be sufficiently informative in most cases,
 				# and in the cases where it isn't, an element traceback should
 				# ultimately yield the pertinent information
-				yield (x[0], repr(data[1])[2:-1])
+				yield (x[0], repr(x[1])[2:-1])
 
 	def decode_notice(self, notice):
 		notice = self._decode_failsafe(notice.items())
@@ -578,10 +648,7 @@ class Output(object):
 		self._pq_cursor_id = self.database.typio.encode(cursor_id)
 		# If the cursor's id was generated, it should be garbage collected.
 		if cursor_id == ID(self):
-			garbage = self.database.pq.garbage_cursors.append
-			cid = self._pq_cursor_id
-			# Callback for closing the cursor on remote end.
-			self._del = wref(self, lambda x: garbage(cid))
+			self.database.pq.register_cursor(self, self._pq_cursor_id)
 		self._quoted_cursor_id = '"' + cursor_id.replace('"', '""') + '"'
 		self._init()
 
@@ -590,11 +657,8 @@ class Output(object):
 
 	def close(self):
 		if self.closed is False:
-			self.database.pq.garbage_cursors.append(self._pq_cursor_id)
+			self.database.pq.trash_cursor(self._pq_cursor_id)
 		self.closed = True
-		# Don't need the weakref anymore.
-		if hasattr(self, '_del'):
-			del self._del
 
 	def _ins(self, *args):
 		return xact.Instruction(*args, asynchook = self.database._receive_async)
@@ -691,12 +755,13 @@ class Output(object):
 		r = range(len(l))
 		try:
 			return [unpack(l[i]) for i in r]
-		except:
-			try:
-				i = next(r)
-			except StopIteration:
-				i = len(l)
-			self._raise_column_tuple_error(self._output_io, (l[i],), 0)
+		except Exception:
+			cause = sys.exc_info()[1]
+		try:
+			i = next(r)
+		except StopIteration:
+			i = len(l)
+		self._raise_column_tuple_error(cause, self._output_io, (l[i],), 0)
 
 	# Process the element.Tuple message in x for rows()
 	def _process_tuple_chunk_Row(self, x,
@@ -712,7 +777,7 @@ class Output(object):
 	def _process_tuple_chunk(self, x, proc = process_chunk):
 		return proc(self._output_io, x, self._raise_column_tuple_error)
 
-	def _raise_column_tuple_error(self, procs, tup, itemnum):
+	def _raise_column_tuple_error(self, cause, procs, tup, itemnum):
 		'for column processing'
 		# The element traceback will include the full list of parameters.
 		data = repr(tup[itemnum])
@@ -735,7 +800,7 @@ class Output(object):
 			(b'H', "Try casting the column to 'text'."),
 			(b'P', str(itemnum)),
 		))
-		self.database.typio.raise_client_error(em, creator = self)
+		self.database.typio.raise_client_error(em, creator = self, cause = cause)
 
 	@property
 	def state(self):
@@ -1063,7 +1128,11 @@ class Cursor(Output, pg_api.Cursor):
 					])
 
 	def __next__(self):
-		return self._fetch(self.direction, 1)
+		result = self._fetch(self.direction, 1)
+		if not result:
+			raise StopIteration
+		else:
+			return result[0]
 
 	def read(self, quantity = None, direction = None):
 		if quantity == 0:
@@ -1090,23 +1159,40 @@ class Cursor(Output, pg_api.Cursor):
 				"unknown whence parameter, %r" %(whence,)
 			)
 		rwhence = rwhence.upper()
-		if self.direction is False:
-			if rwhence == 'RELATIVE':
-				offset = -offset
-			elif rwhence == 'ABSOLUTE':
-				rwhence = 'FROM_END'
-			else:
-				rwhence = 'ABSOLUTE'
 
-		if rwhence == 'RELATIVE':
-			if offset < 0:
+		if offset == 'ALL':
+			if rwhence not in ('BACKWARD', 'FORWARD'):
+				rwhence = 'BACKWARD' if self.direction is False else 'FORWARD'
+		else:
+			if offset < 0 and rwhence == 'BACKWARD':
+				offset = -offset
+				rwhence = 'FORWARD'
+
+			if self.direction is False:
+				if offset == 'ALL' and rwhence != 'FORWARD':
+					rwhence = 'BACKWARD'
+				else:
+					if rwhence == 'RELATIVE':
+						offset = -offset
+					elif rwhence == 'ABSOLUTE':
+						rwhence = 'FROM_END'
+					else:
+						rwhence = 'ABSOLUTE'
+
+		if rwhence in ('RELATIVE', 'BACKWARD', 'FORWARD'):
+			if offset == 'ALL':
 				cmd = self._pq_xp_move(
-					str(-offset).encode('ascii'), b'BACKWARD'
+					str(offset).encode('ascii'), str(rwhence).encode('ascii')
 				)
 			else:
-				cmd = self._pq_xp_move(
-					str(offset).encode('ascii'), b'RELATIVE'
-				)
+				if offset < 0:
+					cmd = self._pq_xp_move(
+						str(-offset).encode('ascii'), b'BACKWARD'
+					)
+				else:
+					cmd = self._pq_xp_move(
+						str(offset).encode('ascii'), str(rwhence).encode('ascii')
+					)
 		elif rwhence == 'ABSOLUTE':
 			cmd = self._pq_xp_move(str(offset).encode('ascii'), b'ABSOLUTE')
 		else:
@@ -1120,6 +1206,45 @@ class Cursor(Output, pg_api.Cursor):
 		self.database._pq_push(x, self)
 		self.database._pq_complete()
 
+		count = None
+		complete = element.Complete.type
+		for cm in x.messages_received():
+			if getattr(cm, 'type', None) == complete:
+				count = cm.extract_count()
+				break
+
+		# XXX: Raise if count is None?
+		return count
+
+class SingleExecution(pg_api.Execution):
+	database = None
+	def __init__(self, database):
+		self._prepare = database.prepare
+
+	def load_rows(self, query, *parameters):
+		return self._prepare(query).load_rows(*parameters)
+
+	def load_chunks(self, query, *parameters):
+		return self._prepare(query).load_chunks(*parameters)
+
+	def __call__(self, query, *parameters):
+		return self._prepare(query)(*parameters)
+
+	def declare(self, query, *parameters):
+		return self._prepare(query).declare(*parameters)
+
+	def rows(self, query, *parameters):
+		return self._prepare(query).rows(*parameters)
+
+	def chunks(self, query, *parameters):
+		return self._prepare(query).chunks(*parameters)
+
+	def column(self, query, *parameters):
+		return self._prepare(query).column(*parameters)
+
+	def first(self, query, *parameters):
+		return self._prepare(query).first(*parameters)
+
 class Statement(pg_api.Statement):
 	string = None
 	database = None
@@ -1134,7 +1259,7 @@ class Statement(pg_api.Statement):
 		yield (None, '[' + self.state + ']')
 		if hasattr(self._xact, 'error_message'):
 			# be very careful not to trigger an exception.
-			# even in the cases of effective protocol errors, 
+			# even in the cases of effective protocol errors,
 			# it is important not to bomb out.
 			pos = self._xact.error_message.get(b'P')
 			if pos is not None and pos.isdigit():
@@ -1182,7 +1307,7 @@ class Statement(pg_api.Statement):
 				yield (
 					'results',
 					'(' + ', '.join([
-						n + ' ' + t for n,t in zip(cn,ct)
+						'{!r} {!r}'.format(n, t) for n,t in zip(cn,ct)
 					]) + ')'
 				)
 			else:
@@ -1208,10 +1333,8 @@ class Statement(pg_api.Statement):
 		self._pq_statement_id = database.typio._encode(self.statement_id)[0]
 
 		if not statement_id:
-			garbage = database.pq.garbage_statements.append
-			sid = self._pq_statement_id
-			# Callback for closing the statement on remote end.
-			self._del = wref(self, lambda x: garbage(sid))
+			# Register statement on a connection to close it automatically on db end
+			database.pq.register_statement(self, self._pq_statement_id)
 
 	def __repr__(self):
 		return '<{mod}.{name}[{ci}] {state}>'.format(
@@ -1231,7 +1354,7 @@ class Statement(pg_api.Statement):
 	# process_tuple failed(exception). The parameters could not be packed.
 	# This function is called with the given information in the context
 	# of the original exception(to allow chaining).
-	def _raise_parameter_tuple_error(self, procs, tup, itemnum):
+	def _raise_parameter_tuple_error(self, cause, procs, tup, itemnum):
 		# Find the SQL type name. This should *not* hit the server.
 		typ = self.database.typio.sql_type_from_oid(
 			self.pg_parameter_types[itemnum]
@@ -1254,11 +1377,11 @@ class Statement(pg_api.Statement):
 			(b'H', "Try casting the parameter to 'text', then to the target type."),
 			(b'P', str(itemnum))
 		))
-		self.database.typio.raise_client_error(em, creator = self)
+		self.database.typio.raise_client_error(em, creator = self, cause = cause)
 
 	##
 	# Similar to the parameter variant.
-	def _raise_column_tuple_error(self, procs, tup, itemnum):
+	def _raise_column_tuple_error(self, cause, procs, tup, itemnum):
 		# Find the SQL type name. This should *not* hit the server.
 		typ = self.database.typio.sql_type_from_oid(
 			self.pg_column_types[itemnum]
@@ -1281,7 +1404,7 @@ class Statement(pg_api.Statement):
 			(b'H', "Try casting the column to 'text'."),
 			(b'P', str(itemnum)),
 		))
-		self.database.typio.raise_client_error(em, creator = self)
+		self.database.typio.raise_client_error(em, creator = self, cause = cause)
 
 	@property
 	def state(self) -> str:
@@ -1352,11 +1475,8 @@ class Statement(pg_api.Statement):
 
 	def close(self):
 		if self.closed is False:
-			self.database.pq.garbage_statements.append(self._pq_statement_id)
+			self.database.pq.trash_statement(self._pq_statement_id)
 		self.closed = True
-		# Don't need the weakref anymore.
-		if hasattr(self, '_del'):
-			del self._del
 
 	def _init(self):
 		"""
@@ -1593,7 +1713,13 @@ class Statement(pg_api.Statement):
 			else:
 				# Probably a Null command.
 				return None
-			return cm.extract_count() or cm.extract_command()
+
+			count = cm.extract_count()
+			if count is None:
+				command = cm.extract_command()
+				if command is not None:
+					return command.decode('ascii')
+			return count
 
 	def _load_copy_chunks(self, chunks, *parameters):
 		"""
@@ -1646,14 +1772,19 @@ class Statement(pg_api.Statement):
 		self.database._pq_complete()
 		self.database.pq.synchronize()
 
-	def _load_tuple_chunks(self, chunks):
+	def _load_tuple_chunks(self, chunks, tuple=tuple):
 		pte = self._raise_parameter_tuple_error
 		last = (element.SynchronizeMessage,)
+
+		Bind = element.Bind
+		Instruction = xact.Instruction
+		Execute = element.Execute
+
 		try:
 			for chunk in chunks:
 				bindings = [
 					(
-						element.Bind(
+						Bind(
 							b'',
 							self._pq_statement_id,
 							self._input_formats,
@@ -1662,13 +1793,13 @@ class Statement(pg_api.Statement):
 							),
 							(),
 						),
-						element.Execute(b'', 1),
+						Execute(b'', 1),
 					)
 					for t in chunk
 				]
 				bindings.append(last)
 				self.database._pq_push(
-					xact.Instruction(
+					Instruction(
 						chain.from_iterable(bindings),
 						asynchook = self.database._receive_async
 					),
@@ -1786,7 +1917,7 @@ class StoredProcedure(pg_api.StoredProcedure):
 				# ($1::type, $2::type, ... $n::type)
 				', '.join([
 					 '$%d::%s' %(x + 1, database.typio.sql_type_from_oid(proargs[x]))
-					 for x in range(len(proargs))
+					 for x in range(0, len(proargs))
 				]),
 				# Description for anonymous record returns
 				(description and \
@@ -1801,9 +1932,6 @@ class SettingsCM(object):
 		self.database = database
 		self.settings_to_set = settings_to_set
 
-	def __context__(self):
-		return self
-
 	def __enter__(self):
 		if hasattr(self, 'stored_settings'):
 			raise RuntimeError("cannot re-use setting CMs")
@@ -1862,7 +1990,7 @@ class Settings(pg_api.Settings):
 		def fget(self):
 			return pg_str.split_ident(self["search_path"])
 		def fset(self, value):
-			self.settings['search_path'] = ','.join([
+			self['search_path'] = ','.join([
 				'"%s"' %(x.replace('"', '""'),) for x in value
 			])
 		def fdel(self):
@@ -1968,24 +2096,19 @@ class Transaction(pg_api.Transaction):
 
 	mode = None
 	isolation = None
-	gid = None
 
-	_e_factors = ('database', 'gid', 'isolation', 'mode')
+	_e_factors = ('database', 'isolation', 'mode')
 
 	def _e_metas(self):
 		yield (None, self.state)
 
-	def __init__(self, database, isolation = None, mode = None, gid = None):
+	def __init__(self, database, isolation = None, mode = None):
 		self.database = database
-		self.gid = gid
 		self.isolation = isolation
 		self.mode = mode
 		self.state = 'initialized'
 		self.type = None
 
-	def __context__(self):
-		return self
-
 	def __enter__(self):
 		self.start()
 		return self
@@ -2011,26 +2134,10 @@ class Transaction(pg_api.Transaction):
 					# If an error occurs, clean up the transaction state
 					# and raise as needed.
 				except pg_exc.ActiveTransactionError as err:
-					##
-					# Failed COMMIT PREPARED <gid>?
-					# Likely cases:
-					#  - User exited block without preparing the transaction.
-					##
-					if not self.database.closed and self.gid is not None:
+					if not self.database.closed:
 						# adjust the state so rollback will do the right thing and abort.
 						self.state = 'open'
 						self.rollback()
-					##
-					# The other exception that *can* occur is
-					# UndefinedObjectError in which:
-					#  - User issued C/RB P <gid> before exit, but not via xact methods.
-					#  - User adjusted gid after prepare().
-					#
-					# But the occurrence of this exception means it's not in an active
-					# transaction, which means no cleanup other than raise is necessary.
-					err.details['cause'] = \
-						"The prepared transaction was not " \
-						"prepared prior to the block's exit."
 					raise
 		elif issubclass(typ, Exception):
 			# There's an exception, so only rollback if the connection
@@ -2076,7 +2183,7 @@ class Transaction(pg_api.Transaction):
 			)
 		else:
 			self.type = 'savepoint'
-			if (self.gid, self.isolation, self.mode) != (None,None,None):
+			if (self.isolation, self.mode) != (None,None):
 				em = element.ClientError((
 					(b'S', 'ERROR'),
 					(b'C', '--OPE'),
@@ -2089,60 +2196,15 @@ class Transaction(pg_api.Transaction):
 		self.state = 'open'
 	begin = start
 
-	@staticmethod
-	def _prepare_string(id):
-		"2pc prepared transaction 'gid'"
-		return "PREPARE TRANSACTION '" + id.replace("'", "''") + "';"
-
 	@staticmethod
 	def _release_string(id):
 		'release "";'
 		return 'RELEASE "xact(' + id.replace('"', '""') + ')";'
 
-	def prepare(self):
-		if self.state == 'prepared':
-			return
-		if self.state != 'open':
-			em = element.ClientError((
-				(b'S', 'ERROR'),
-				(b'C', '--OPE'),
-				(b'M', "transaction state must be 'open' in order to prepare"),
-			))
-			self.database.typio.raise_client_error(em, creator = self)
-		if self.type != 'block':
-			em = element.ClientError((
-				(b'S', 'ERROR'),
-				(b'C', '--OPE'),
-				(b'M', "improper transaction type to prepare"),
-			))
-			self.database.typio.raise_client_error(em, creator = self)
-		q = self._prepare_string(self.gid)
-		self.database.execute(q)
-		self.state = 'prepared'
-
-	def recover(self):
-		if self.state != 'initialized':
-			em = element.ClientError((
-				(b'S', 'ERROR'),
-				(b'C', '--OPE'),
-				(b'M', "improper state for prepared transaction recovery"),
-			))
-			self.database.typio.raise_client_error(em, creator = self)
-		if self.database.sys.xact_is_prepared(self.gid):
-			self.state = 'prepared'
-			self.type = 'block'
-		else:
-			em = element.ClientError((
-				(b'S', 'ERROR'),
-				(b'C', '42704'), # UndefinedObjectError
-				(b'M', "prepared transaction does not exist"),
-			))
-			self.database.typio.raise_client_error(em, creator = self)
-
 	def commit(self):
 		if self.state == 'committed':
 			return
-		if self.state not in ('prepared', 'open'):
+		if self.state != 'open':
 			em = element.ClientError((
 				(b'S', 'ERROR'),
 				(b'C', '--OPE'),
@@ -2151,26 +2213,15 @@ class Transaction(pg_api.Transaction):
 			self.database.typio.raise_client_error(em, creator = self)
 
 		if self.type == 'block':
-			if self.gid is not None:
-				# User better have prepared it.
-				q = "COMMIT PREPARED '" + self.gid.replace("'", "''") + "';"
-			else:
-				q = 'COMMIT'
+			q = 'COMMIT'
 		else:
-			if self.gid is not None:
-				em = element.ClientError((
-					(b'S', 'ERROR'),
-					(b'C', '--OPE'),
-					(b'M', "savepoint configured with global identifier"),
-				))
-				self.database.typio.raise_client_error(em, creator = self)
 			q = self._release_string(hex(id(self)))
 		self.database.execute(q)
 		self.state = 'committed'
 
 	@staticmethod
-	def _rollback_to_string(id):
-		return 'ROLLBACK TO "xact(' + id.replace('"', '""') + ')";'
+	def _rollback_to_string(id, fmt = 'ROLLBACK TO "xact({0})"; RELEASE "xact({0})";'.format):
+		return fmt(id.replace('"', '""'))
 
 	def rollback(self):
 		if self.state == 'aborted':
@@ -2184,10 +2235,7 @@ class Transaction(pg_api.Transaction):
 			self.database.typio.raise_client_error(em, creator = self)
 
 		if self.type == 'block':
-			if self.state == 'prepared':
-				q = "ROLLBACK PREPARED '" + self.gid.replace("'", "''") + "'"
-			else:
-				q = 'ABORT;'
+			q = 'ABORT;'
 		elif self.type == 'savepoint':
 			q = self._rollback_to_string(hex(id(self)))
 		else:
@@ -2245,9 +2293,6 @@ class Connection(pg_api.Connection):
 			self.closed and 'closed' or '%s' %(self.pq.state,)
 		)
 
-	def __context__(self):
-		return self
-
 	def __exit__(self, type, value, tb):
 		# Don't bother closing unless it's a normal exception.
 		if type is None or issubclass(type, Exception):
@@ -2272,19 +2317,24 @@ class Connection(pg_api.Connection):
 		sql = "DO " + qlit(source) + " LANGUAGE " + qid(language) + ";"
 		self.execute(sql)
 
-	def xact(self, gid = None, isolation = None, mode = None):
-		x = Transaction(self, gid = gid, isolation = isolation, mode = mode)
+	def xact(self, isolation = None, mode = None):
+		x = Transaction(self, isolation = isolation, mode = mode)
 		return x
 
 	def prepare(self,
 		sql_statement_string : str,
 		statement_id = None,
+		Class = Statement
 	) -> Statement:
-		ps = Statement(self, statement_id, sql_statement_string)
+		ps = Class(self, statement_id, sql_statement_string)
 		ps._init()
 		ps._fini()
 		return ps
 
+	@property
+	def query(self, Class = SingleExecution):
+		return Class(self)
+
 	def statement_from_id(self, statement_id : str) -> Statement:
 		ps = Statement(self, statement_id, None)
 		ps._init()
@@ -2302,26 +2352,36 @@ class Connection(pg_api.Connection):
 
 	@property
 	def closed(self) -> bool:
-		if not hasattr(self, 'pq'):
+		if getattr(self, 'pq', None) is None:
 			return True
 		if hasattr(self.pq, 'socket') and self.pq.xact is not None:
 			return self.pq.xact.fatal is True
 		return False
 
-	def close(self):
-		if self.closed:
-			return
+	def close(self, getattr = getattr):
 		# Write out the disconnect message if the socket is around.
 		# If the connection is known to be lost, don't bother. It will
 		# generate an extra exception.
-		x = self.pq.xact
-		if x:
-			# don't raise?
+		if getattr(self, 'pq', None) is None or getattr(self.pq, 'socket', None) is None:
+			# No action to take.
+			return
+
+		x = getattr(self.pq, 'xact', None)
+		if x is not None and x.fatal is not True:
+			# finish the existing pq transaction iff it's not Closing.
 			self.pq.complete()
-			if self.closed:
-				return
-		self.pq.push(xact.Closing())
-		self.pq.complete()
+
+		if self.pq.xact is None:
+			# It completed the existing transaction.
+			self.pq.push(xact.Closing())
+			self.pq.complete()
+			if self.pq.socket:
+				self.pq.complete()
+
+		# Close the socket if there is one.
+		if self.pq.socket:
+			self.pq.socket.close()
+			self.pq.socket = None
 
 	@property
 	def state(self) -> str:
@@ -2356,16 +2416,25 @@ class Connection(pg_api.Connection):
 		if self.closed is False:
 			# already connected? just return.
 			return
-		# It's closed.
 
 		if hasattr(self, 'pq'):
 			# It's closed, *but* there's a PQ connection..
 			x = self.pq.xact
 			self.typio.raise_error(x.error_message, cause = getattr(x, 'exception', None), creator = self)
 
+		# It's closed.
+		try:
+			self._establish()
+		except Exception:
+			# Close it up on failure.
+			self.close()
+			raise
+
+	def _establish(self):
+		# guts of connect()
 		self.pq = None
 		# if any exception occurs past this point, the connection
-		# will not be usable.
+		# object will not be usable.
 		timeout = self.connector.connect_timeout
 		sslmode = self.connector.sslmode or 'prefer'
 		failures = []
@@ -2440,7 +2509,13 @@ class Connection(pg_api.Connection):
 						self._receive_async(x)
 				# success!
 				break
+			elif pq.socket is not None:
+				# In this case, an application/protocol error occurred.
+				# Close out the sockets ourselves.
+				pq.socket.close()
 
+			# Identify whether or not we can skip the attempt.
+			# Whether or not we can skip depends entirely on the SSL parameter.
 			if sslmode == 'prefer' and ssl is False and didssl is False:
 				# In this case, the server doesn't support SSL or it's
 				# turned off. Therefore, the "without_ssl" attempt need
@@ -2488,6 +2563,8 @@ class Connection(pg_api.Connection):
 		vi = self.version_info[:2]
 		if vi <= (8,1):
 			sd = self.sys.startup_data_only_version()
+		elif vi >= (9,2):
+			sd = self.sys.startup_data_92()
 		else:
 			sd = self.sys.startup_data()
 		# connection info
@@ -2639,7 +2716,8 @@ class Connection(pg_api.Connection):
 		self.connector = connector
 		# raw notify messages
 		self._notifies = []
-		self.typio = TypeIO(self)
+		self.fileno = -1
+		self.typio = self.connector.driver.typio(self)
 		self.typio.set_encoding('ascii')
 		self.settings = Settings(self)
 # class Connection
@@ -2681,7 +2759,7 @@ class Connector(pg_api.Connector):
 		"""
 		Generate a list of callables that will be used to attempt to make the
 		connection to the server. It is assumed that each factory will produce
-		an object with a socket interface that is ready for reading and writing 
+		an object with a socket interface that is ready for reading and writing
 		data.
 
 		The callables in the sequence must take a timeout parameter.
@@ -2739,15 +2817,17 @@ class Connector(pg_api.Connector):
 		##
 		# Attempt to accommodate for literal treatment of startup data.
 		##
-		self._startup_parameters = {
+		self._startup_parameters = tuple([
 			# All keys go in utf-8. However, ascii would probably be good enough.
-			k.encode('utf-8') : \
+			(
+				k.encode('utf-8'),
 			# If it's a str(), encode in the hinted server_encoding.
 			# Otherwise, convert the object(int, float, bool, etc) into a string
 			# and treat it as utf-8.
-			v.encode(se) if type(v) is str else str(v).encode('utf-8')
+				v.encode(se) if type(v) is str else str(v).encode('utf-8')
+			)
 			for k, v in tnkw.items()
-		}
+		])
 		self._password = (self.password or '').encode(se)
 		self._socket_secure = {
 			'keyfile' : self.sslkeyfile,
@@ -2765,41 +2845,49 @@ class SocketConnector(Connector):
 		to the target host.
 		"""
 
-class IP4(SocketConnector):
-	'Connector for establishing IPv4 connections'
-	ipv = 4
+	def create_socket_factory(self, **params):
+		return SocketFactory(**params)
+
+class IPConnector(SocketConnector):
 	def socket_factory_sequence(self):
 		return self._socketcreators
 
+	def socket_factory_params(self, host, port, ipv, **kw):
+		if ipv != self.ipv:
+			raise TypeError("'ipv' keyword must be '%d'" % self.ipv)
+		if host is None:
+			raise TypeError("'host' is a required keyword and cannot be 'None'")
+		if port is None:
+			raise TypeError("'port' is a required keyword and cannot be 'None'")
+
+		return {'socket_create': (self.address_family, socket.SOCK_STREAM),
+				'socket_connect': (host, int(port))}
+
+	def __init__(self, host, port, ipv, **kw):
+		params = self.socket_factory_params(host, port, ipv, **kw)
+		self.host, self.port = params['socket_connect']
+		# constant socket connector
+		self._socketcreator = self.create_socket_factory(**params)
+		self._socketcreators = (self._socketcreator,)
+		super().__init__(**kw)
+
+class IP4(IPConnector):
+	'Connector for establishing IPv4 connections'
+	ipv = 4
+	address_family = socket.AF_INET
+
 	def __init__(self,
 		host : "IPv4 Address (str)" = None,
 		port : int = None,
 		ipv = 4,
 		**kw
 	):
-		if ipv != self.ipv:
-			raise TypeError("'ipv' keyword must be '4'")
-		if host is None:
-			raise TypeError("'host' is a required keyword and cannot be 'None'")
-		if port is None:
-			raise TypeError("'port' is a required keyword and cannot be 'None'")
-		self.host = host
-		self.port = int(port)
-		# constant socket connector
-		self._socketcreator = SocketFactory(
-			(socket.AF_INET, socket.SOCK_STREAM),
-			(self.host, self.port)
-		)
-		self._socketcreators = (
-			self._socketcreator,
-		)
-		super().__init__(**kw)
+		super().__init__(host, port, ipv, **kw)
 
-class IP6(SocketConnector):
+class IP6(IPConnector):
 	'Connector for establishing IPv6 connections'
 	ipv = 6
-	def socket_factory_sequence(self):
-		return self._socketcreators
+	address_family = socket.AF_INET6
 
 	def __init__(self,
 		host : "IPv6 Address (str)" = None,
@@ -2807,37 +2895,25 @@ class IP6(SocketConnector):
 		ipv = 6,
 		**kw
 	):
-		if ipv != self.ipv:
-			raise TypeError("'ipv' keyword must be '6'")
-		if host is None:
-			raise TypeError("'host' is a required keyword and cannot be 'None'")
-		if port is None:
-			raise TypeError("'port' is a required keyword and cannot be 'None'")
-		self.host = host
-		self.port = int(port)
-		# constant socket connector
-		self._socketcreator = SocketFactory(
-			(socket.AF_INET6, socket.SOCK_STREAM),
-			(self.host, self.port)
-		)
-		self._socketcreators = (
-			self._socketcreator,
-		)
-		super().__init__(**kw)
+		super().__init__(host, port, ipv, **kw)
 
 class Unix(SocketConnector):
 	'Connector for establishing unix domain socket connections'
 	def socket_factory_sequence(self):
 		return self._socketcreators
 
-	def __init__(self, unix = None, **kw):
+	def socket_factory_params(self, unix):
 		if unix is None:
 			raise TypeError("'unix' is a required keyword and cannot be 'None'")
-		self.unix = unix
+
+		return {'socket_create': (socket.AF_UNIX, socket.SOCK_STREAM),
+				'socket_connect': unix}
+
+	def __init__(self, unix = None, **kw):
+		params = self.socket_factory_params(unix)
+		self.unix = params['socket_connect']
 		# constant socket connector
-		self._socketcreator = SocketFactory(
-			(socket.AF_UNIX, socket.SOCK_STREAM), self.unix
-		)
+		self._socketcreator = self.create_socket_factory(**params)
 		self._socketcreators = (self._socketcreator,)
 		super().__init__(**kw)
 
@@ -2854,12 +2930,18 @@ class Host(SocketConnector):
 		"""
 		return [
 			# (AF, socktype, proto), (IP, Port)
-			SocketFactory(x[0:3], x[4][:2], self._socket_secure)
+			self.create_socket_factory(**(self.socket_factory_params(x[0:3], x[4][:2],
+																	self._socket_secure)))
 			for x in socket.getaddrinfo(
 				self.host, self.port, self._address_family, socket.SOCK_STREAM
 			)
 		]
 
+	def socket_factory_params(self, socktype, address, sslparams):
+		return {'socket_create': socktype,
+				'socket_connect': address,
+				'socket_secure': sslparams}
+
 	def __init__(self,
 		host : str = None,
 		port : (str, int) = None,
