@@ -1,6 +1,5 @@
 ##
-# copyright 2009, James William Pye
-# http://python.projects.postgresql.org
+# .iri
 ##
 """
 Parse and serialize PQ IRIs.
@@ -47,7 +46,7 @@ def structure(d, fieldproc = ri.unescape):
 			if host.startswith('unix:'):
 				cpd['unix'] = host[len('unix:'):].replace(':','/')
 			else:
-				cpd['host'] = host[1:-1]
+				cpd['host'] = host
 		else:
 			cpd['host'] = fieldproc(host)
 
