@@ -29,11 +29,21 @@ class test_cluster(unittest.TestCase):
 
 	def init(self, *args, **kw):
 		self.cluster.init(*args, **kw)
+
+		vi = self.cluster.installation.version_info[:2]
+		if vi >= (9, 3):
+			usd = 'unix_socket_directories'
+		else:
+			usd = 'unix_socket_directory'
+
+		if vi > (9, 6):
+			self.cluster.settings['max_wal_senders'] = '0'
+
 		self.cluster.settings.update({
 			'max_connections' : '8',
 			'listen_addresses' : 'localhost',
 			'port' : '6543',
-			'silent_mode' : 'off',
+			usd : self.cluster.data_directory,
 		})
 
 	def testSilentMode(self):
@@ -46,11 +56,15 @@ class test_cluster(unittest.TestCase):
 			# silent_mode is not supported on windows by PG.
 			if sys.platform in ('win32','win64'):
 				pass
+			elif self.cluster.installation.version_info[:2] >= (9, 2):
+				pass
 			else:
 				raise
 		else:
 			if sys.platform in ('win32','win64'):
-				self.fail("silent_mode supported on windows")
+				self.fail("silent_mode unexpectedly supported on windows")
+			elif self.cluster.installation.version_info[:2] >= (9, 2):
+				self.fail("silent_mode unexpectedly supported on PostgreSQL >=9.2")
 
 	def testSuperPassword(self):
 		self.init(
@@ -65,7 +79,7 @@ class test_cluster(unittest.TestCase):
 			database='template1',
 		)
 		with c:
-			self.failUnless(c.prepare('select 1').first() == 1)
+			self.assertEqual(c.prepare('select 1').first(), 1)
 
 	def testNoParameters(self):
 		'simple init and drop'
