@@ -4,17 +4,158 @@
 import sys
 import os
 import unittest
+import atexit
+import socket
+import errno
 
+from ..python.socket import find_available_port
+
+from .. import installation
+from .. import cluster as pg_cluster
 from .. import exceptions as pg_exc
-from .. import unittest as pg_unittest
 
 from ..driver import dbapi20 as dbapi20
 from .. import driver as pg_driver
 from .. import open as pg_open
 
+
+def check_for_ipv6():
+	result = False
+	if socket.has_ipv6:
+		try:
+			socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
+			result = True
+		except socket.error as e:
+			errs = [errno.EAFNOSUPPORT]
+			WSAEAFNOSUPPORT = getattr(errno, 'WSAEAFNOSUPPORT', None)
+			if WSAEAFNOSUPPORT is not None:
+				errs.append(WSAEAFNOSUPPORT)
+			if e.errno not in errs:
+				raise
+	return result
+
+
 msw = sys.platform in ('win32', 'win64')
 
-class test_connect(pg_unittest.TestCaseWithCluster):
+# win32 binaries don't appear to be built with ipv6
+has_ipv6 = check_for_ipv6() and not msw
+
+has_unix_sock = not msw
+
+
+class TestCaseWithCluster(unittest.TestCase):
+	"""
+	postgresql.driver *interface* tests.
+	"""
+	def __init__(self, *args, **kw):
+		super().__init__(*args, **kw)
+		self.installation = installation.default()
+		self.cluster_path = \
+			'pypg_test_' \
+			+ str(os.getpid()) + getattr(self, 'cluster_path_suffix', '')
+
+		if self.installation is None:
+			sys.stderr.write("ERROR: cannot find 'default' pg_config\n")
+			sys.stderr.write(
+				"HINT: set the PGINSTALLATION environment variable to the `pg_config` path\n"
+			)
+			sys.exit(1)
+
+		self.cluster = pg_cluster.Cluster(
+			self.installation,
+			self.cluster_path,
+		)
+		if self.cluster.initialized():
+			self.cluster.drop()
+
+		self.disable_replication = self.installation.version_info[:2] > (9, 6)
+
+	def configure_cluster(self):
+		self.cluster_port = find_available_port()
+		if self.cluster_port is None:
+			pg_exc.ClusterError(
+				'failed to find a port for the test cluster on localhost',
+				creator = self.cluster
+			).raise_exception()
+
+		listen_addresses = '127.0.0.1'
+		if has_ipv6:
+			listen_addresses += ',::1'
+
+		self.cluster.settings.update(dict(
+			port = str(self.cluster_port),
+			max_connections = '6',
+			shared_buffers = '24',
+			listen_addresses = listen_addresses,
+			log_destination = 'stderr',
+			log_min_messages = 'FATAL',
+		))
+
+		if self.disable_replication:
+			self.cluster.settings.update({
+				'max_wal_senders': '0',
+			})
+
+		if self.cluster.installation.version_info[:2] < (9, 3):
+			self.cluster.settings.update(dict(
+				unix_socket_directory = self.cluster.data_directory,
+			))
+		else:
+			self.cluster.settings.update(dict(
+				unix_socket_directories = self.cluster.data_directory,
+			))
+
+		# 8.4 turns prepared transactions off by default.
+		if self.cluster.installation.version_info >= (8,1):
+			self.cluster.settings.update(dict(
+				max_prepared_transactions = '3',
+			))
+
+	def initialize_database(self):
+		c = self.cluster.connection(
+			user = 'test',
+			database = 'template1',
+		)
+		with c:
+			if c.prepare(
+				"select true from pg_catalog.pg_database " \
+				"where datname = 'test'"
+			).first() is None:
+				c.execute('create database test')
+
+	def connection(self, *args, **kw):
+		return self.cluster.connection(*args, user = 'test', **kw)
+
+	def run(self, *args, **kw):
+		if not self.cluster.initialized():
+			self.cluster.encoding = 'utf-8'
+			self.cluster.init(
+				user = 'test',
+				encoding = self.cluster.encoding,
+				logfile = None,
+			)
+			sys.stderr.write('*')
+			try:
+				atexit.register(self.cluster.drop)
+				self.configure_cluster()
+				self.cluster.start(logfile = sys.stdout)
+				self.cluster.wait_until_started()
+				self.initialize_database()
+			except Exception:
+				self.cluster.drop()
+				atexit.unregister(self.cluster.drop)
+				raise
+		if not self.cluster.running():
+			self.cluster.start()
+			self.cluster.wait_until_started()
+
+		db = self.connection()
+		with db:
+			self.db = db
+			return super().run(*args, **kw)
+			self.db = None
+
+class test_connect(TestCaseWithCluster):
 	"""
 	postgresql.driver connectivity tests
 	"""
@@ -24,25 +165,35 @@ class test_connect(pg_unittest.TestCaseWithCluster):
 	params = {}
 	cluster_path_suffix = '_test_connect'
 
+	mk_common_users = """
+		CREATE USER md5 WITH ENCRYPTED PASSWORD 'md5_password';
+		CREATE USER password WITH ENCRYPTED PASSWORD 'password_password';
+		CREATE USER trusted;
+	"""
+
+	mk_crypt_user = """
+		-- crypt doesn't work with encrypted passwords:
+		-- http://www.postgresql.org/docs/8.2/interactive/auth-methods.html#AUTH-PASSWORD
+		CREATE USER crypt WITH UNENCRYPTED PASSWORD 'crypt_password';
+	"""
+
 	def __init__(self, *args, **kw):
 		super().__init__(*args,**kw)
 		# 8.4 nixed this.
-		self.do_crypt = self.cluster.installation.version_info < (8,4)
-		self.do_unix = sys.platform != msw
+		vi = self.cluster.installation.version_info
+		self.check_crypt_user = (vi < (8,4))
 
 	def configure_cluster(self):
 		super().configure_cluster()
-		self.cluster.settings.update({
-			'log_min_messages' : 'log',
-			'unix_socket_directory' : self.cluster.data_directory,
-		})
+		self.cluster.settings['log_min_messages'] = 'log'
 
 		# Configure the hba file with the supported methods.
 		with open(self.cluster.hba_file, 'w') as hba:
 			hosts = ['0.0.0.0/0',]
-			if not msw:
+			if has_ipv6:
 				hosts.append('0::0/0')
-			methods = ['md5', 'password'] + (['crypt'] if self.do_crypt else [])
+
+			methods = ['md5', 'password'] + (['crypt'] if self.check_crypt_user else [])
 			for h in hosts:
 				for m in methods:
 					# user and method are the same name.
@@ -50,38 +201,24 @@ class test_connect(pg_unittest.TestCaseWithCluster):
 						h = h,
 						m = m
 					)])
+
 			# trusted
 			hba.writelines(["local all all trust\n"])
 			hba.writelines(["host test trusted 0.0.0.0/0 trust\n"])
-			if not msw:
+			if has_ipv6:
 				hba.writelines(["host test trusted 0::0/0 trust\n"])
 			# admin lines
 			hba.writelines(["host all test 0.0.0.0/0 trust\n"])
-			if not msw:
+			if has_ipv6:
 				hba.writelines(["host all test 0::0/0 trust\n"])
 
 	def initialize_database(self):
 		super().initialize_database()
+
 		with self.cluster.connection(user = 'test') as db:
-			db.execute(
-				"""
-CREATE USER md5 WITH
-	ENCRYPTED PASSWORD 'md5_password'
-;
-
--- crypt doesn't work with encrypted passwords:
--- http://www.postgresql.org/docs/8.2/interactive/auth-methods.html#AUTH-PASSWORD
-CREATE USER crypt WITH
-	UNENCRYPTED PASSWORD 'crypt_password'
-;
-
-CREATE USER password WITH
-	ENCRYPTED PASSWORD 'password_password'
-;
-
-CREATE USER trusted;
-				"""
-			)
+			db.execute(self.mk_common_users)
+			if self.check_crypt_user:
+				db.execute(self.mk_crypt_user)
 
 	def test_pg_open_SQL_ASCII(self):
 		# postgresql.open
@@ -91,9 +228,9 @@ CREATE USER trusted;
 			'pq://' + 'md5:' + 'md5_password@' + host + ':' + str(port) \
 			+ '/test?client_encoding=SQL_ASCII'
 		) as db:
-			self.failUnlessEqual(db.prepare('select 1')(), [(1,)])
-			self.failUnlessEqual(db.settings['client_encoding'], 'SQL_ASCII')
-		self.failUnless(db.closed)
+			self.assertEqual(db.prepare('select 1')(), [(1,)])
+			self.assertEqual(db.settings['client_encoding'], 'SQL_ASCII')
+		self.assertTrue(db.closed)
 
 	def test_pg_open_keywords(self):
 		host, port = self.cluster.address()
@@ -105,8 +242,8 @@ CREATE USER trusted;
 			port = port,
 			database = 'test'
 		) as db:
-			self.failUnlessEqual(db.prepare('select 1')(), [(1,)])
-		self.failUnless(db.closed)
+			self.assertEqual(db.prepare('select 1')(), [(1,)])
+		self.assertTrue(db.closed)
 		# composite test
 		with pg_open(
 			"pq://md5:md5_password@",
@@ -114,7 +251,7 @@ CREATE USER trusted;
 			port = port,
 			database = 'test'
 		) as db:
-			self.failUnlessEqual(db.prepare('select 1')(), [(1,)])
+			self.assertEqual(db.prepare('select 1')(), [(1,)])
 		# override test
 		with pg_open(
 			"pq://md5:foobar@",
@@ -123,7 +260,7 @@ CREATE USER trusted;
 			port = port,
 			database = 'test'
 		) as db:
-			self.failUnlessEqual(db.prepare('select 1')(), [(1,)])
+			self.assertEqual(db.prepare('select 1')(), [(1,)])
 		# and, one with some settings
 		with pg_open(
 			"pq://md5:foobar@?search_path=ieeee",
@@ -133,8 +270,8 @@ CREATE USER trusted;
 			database = 'test',
 			settings = {'search_path' : 'public'}
 		) as db:
-			self.failUnlessEqual(db.prepare('select 1')(), [(1,)])
-			self.failUnlessEqual(db.settings['search_path'], 'public')
+			self.assertEqual(db.prepare('select 1')(), [(1,)])
+			self.assertEqual(db.settings['search_path'], 'public')
 
 	def test_pg_open(self):
 		# postgresql.open
@@ -144,21 +281,21 @@ CREATE USER trusted;
 			'pq://' + 'md5:' + 'md5_password@' + host + ':' + str(port) \
 			+ '/test'
 		) as db:
-			self.failUnlessEqual(db.prepare('select 1')(), [(1,)])
-		self.failUnless(db.closed)
+			self.assertEqual(db.prepare('select 1')(), [(1,)])
+		self.assertTrue(db.closed)
 
 		with pg_open(
 			'pq://' + 'password:' + 'password_password@' + host + ':' + str(port) \
 			+ '/test'
 		) as db:
-			self.failUnlessEqual(db.prepare('select 1')(), [(1,)])
-		self.failUnless(db.closed)
+			self.assertEqual(db.prepare('select 1')(), [(1,)])
+		self.assertTrue(db.closed)
 
 		with pg_open(
 			'pq://' + 'trusted@' + host + ':' + str(port) + '/test'
 		) as db:
-			self.failUnlessEqual(db.prepare('select 1')(), [(1,)])
-		self.failUnless(db.closed)
+			self.assertEqual(db.prepare('select 1')(), [(1,)])
+		self.assertTrue(db.closed)
 
 		# test environment collection
 		pgenv = ('PGUSER', 'PGPORT', 'PGHOST', 'PGSERVICE', 'PGPASSWORD', 'PGDATABASE')
@@ -172,9 +309,9 @@ CREATE USER trusted;
 			os.environ['PGDATABASE'] = 'test'
 			# No arguments, the environment provided everything.
 			with pg_open() as db:
-				self.failUnlessEqual(db.prepare('select 1')(), [(1,)])
-				self.failUnlessEqual(db.prepare('select current_user').first(), 'md5')
-			self.failUnless(db.closed)
+				self.assertEqual(db.prepare('select 1')(), [(1,)])
+				self.assertEqual(db.prepare('select current_user').first(), 'md5')
+			self.assertTrue(db.closed)
 		finally:
 			i = 0
 			for x in stored:
@@ -202,9 +339,9 @@ search_path = public
 					os.environ['PGSERVICE'] = 'myserv'
 					os.environ['PGSYSCONFDIR'] = os.getcwd()
 					with pg_open() as db:
-						self.failUnlessEqual(db.prepare('select 1')(), [(1,)])
-						self.failUnlessEqual(db.prepare('select current_user').first(), 'password')
-						self.failUnlessEqual(db.settings['search_path'], 'public')
+						self.assertEqual(db.prepare('select 1')(), [(1,)])
+						self.assertEqual(db.prepare('select current_user').first(), 'password')
+						self.assertEqual(db.settings['search_path'], 'public')
 				finally:
 					if oldservice is None:
 						os.environ.pop('PGSERVICE', None)
@@ -227,13 +364,13 @@ search_path = public
 			host = host, port = port,
 			**self.params
 		)
-		self.failUnlessEqual(MD5.cursor().execute('select 1').fetchone()[0], 1)
+		self.assertEqual(MD5.cursor().execute('select 1').fetchone()[0], 1)
 		MD5.close()
-		self.failUnlessRaises(pg_exc.ConnectionDoesNotExistError,
+		self.assertRaises(pg_exc.ConnectionDoesNotExistError,
 			MD5.cursor().execute, 'select 1'
 		)
 
-		if self.do_crypt:
+		if self.check_crypt_user:
 			CRYPT = dbapi20.connect(
 				user = 'crypt',
 				database = 'test',
@@ -241,9 +378,9 @@ search_path = public
 				host = host, port = port,
 				**self.params
 			)
-			self.failUnlessEqual(CRYPT.cursor().execute('select 1').fetchone()[0], 1)
+			self.assertEqual(CRYPT.cursor().execute('select 1').fetchone()[0], 1)
 			CRYPT.close()
-			self.failUnlessRaises(pg_exc.ConnectionDoesNotExistError,
+			self.assertRaises(pg_exc.ConnectionDoesNotExistError,
 				CRYPT.cursor().execute, 'select 1'
 			)
 
@@ -254,9 +391,9 @@ search_path = public
 			host = host, port = port,
 			**self.params
 		)
-		self.failUnlessEqual(PASSWORD.cursor().execute('select 1').fetchone()[0], 1)
+		self.assertEqual(PASSWORD.cursor().execute('select 1').fetchone()[0], 1)
 		PASSWORD.close()
-		self.failUnlessRaises(pg_exc.ConnectionDoesNotExistError,
+		self.assertRaises(pg_exc.ConnectionDoesNotExistError,
 			PASSWORD.cursor().execute, 'select 1'
 		)
 
@@ -267,12 +404,23 @@ search_path = public
 			host = host, port = port,
 			**self.params
 		)
-		self.failUnlessEqual(TRUST.cursor().execute('select 1').fetchone()[0], 1)
+		self.assertEqual(TRUST.cursor().execute('select 1').fetchone()[0], 1)
 		TRUST.close()
-		self.failUnlessRaises(pg_exc.ConnectionDoesNotExistError,
+		self.assertRaises(pg_exc.ConnectionDoesNotExistError,
 			TRUST.cursor().execute, 'select 1'
 		)
 
+	def test_dbapi_connect_failure(self):
+		host, port = self.cluster.address()
+		badlogin = (lambda: dbapi20.connect(
+			user = '--',
+			database = '--',
+			password = '...',
+			host = host, port = port,
+			**self.params
+		))
+		self.assertRaises(pg_exc.ClientCannotConnectError, badlogin)
+
 	def test_IP4_connect(self):
 		C = pg_driver.default.ip4(
 			user = 'test',
@@ -282,11 +430,9 @@ search_path = public
 			**self.params
 		)
 		with C() as c:
-			self.failUnlessEqual(c.prepare('select 1').first(), 1)
+			self.assertEqual(c.prepare('select 1').first(), 1)
 
-	if not msw:
-		# win32 binaries don't appear to be built with ipv6
-		# so filter this test on windows.
+	if has_ipv6:
 		def test_IP6_connect(self):
 			C = pg_driver.default.ip6(
 				user = 'test',
@@ -296,7 +442,7 @@ search_path = public
 				**self.params
 			)
 			with C() as c:
-				self.failUnlessEqual(c.prepare('select 1').first(), 1)
+				self.assertEqual(c.prepare('select 1').first(), 1)
 
 	def test_Host_connect(self):
 		C = pg_driver.default.host(
@@ -307,7 +453,7 @@ search_path = public
 			**self.params
 		)
 		with C() as c:
-			self.failUnlessEqual(c.prepare('select 1').first(), 1)
+			self.assertEqual(c.prepare('select 1').first(), 1)
 
 	def test_md5_connect(self):
 		c = self.cluster.connection(
@@ -317,10 +463,10 @@ search_path = public
 			**self.params
 		)
 		with c:
-			self.failUnlessEqual(c.prepare('select current_user').first(), 'md5')
+			self.assertEqual(c.prepare('select current_user').first(), 'md5')
 
 	def test_crypt_connect(self):
-		if self.do_crypt:
+		if self.check_crypt_user:
 			c = self.cluster.connection(
 				user = 'crypt',
 				password = 'crypt_password',
@@ -328,7 +474,7 @@ search_path = public
 				**self.params
 			)
 			with c:
-				self.failUnlessEqual(c.prepare('select current_user').first(), 'crypt')
+				self.assertEqual(c.prepare('select current_user').first(), 'crypt')
 
 	def test_password_connect(self):
 		c = self.cluster.connection(
@@ -337,7 +483,7 @@ search_path = public
 			database = 'test',
 		)
 		with c:
-			self.failUnlessEqual(c.prepare('select current_user').first(), 'password')
+			self.assertEqual(c.prepare('select current_user').first(), 'password')
 
 	def test_trusted_connect(self):
 		c = self.cluster.connection(
@@ -347,10 +493,10 @@ search_path = public
 			**self.params
 		)
 		with c:
-			self.failUnlessEqual(c.prepare('select current_user').first(), 'trusted')
+			self.assertEqual(c.prepare('select current_user').first(), 'trusted')
 
 	def test_Unix_connect(self):
-		if msw:
+		if not has_unix_sock:
 			return
 		unix_domain_socket = os.path.join(
 			self.cluster.data_directory,
@@ -361,22 +507,22 @@ search_path = public
 			unix = unix_domain_socket,
 		)
 		with C() as c:
-			self.failUnlessEqual(c.prepare('select 1').first(), 1)
-			self.failUnlessEqual(c.client_address, None)
+			self.assertEqual(c.prepare('select 1').first(), 1)
+			self.assertEqual(c.client_address, None)
 
 	def test_pg_open_unix(self):
-		if msw:
+		if not has_unix_sock:
 			return
 		unix_domain_socket = os.path.join(
 			self.cluster.data_directory,
 			'.s.PGSQL.' + self.cluster.settings['port']
 		)
 		with pg_open(unix = unix_domain_socket, user = 'test') as c:
-			self.failUnlessEqual(c.prepare('select 1').first(), 1)
-			self.failUnlessEqual(c.client_address, None)
+			self.assertEqual(c.prepare('select 1').first(), 1)
+			self.assertEqual(c.client_address, None)
 		with pg_open('pq://test@[unix:' + unix_domain_socket.replace('/',':') + ']') as c:
-			self.failUnlessEqual(c.prepare('select 1').first(), 1)
-			self.failUnlessEqual(c.client_address, None)
+			self.assertEqual(c.prepare('select 1').first(), 1)
+			self.assertEqual(c.client_address, None)
 
 if __name__ == '__main__':
 	unittest.main()
