@@ -54,8 +54,8 @@ class test_dbapi20(unittest.TestCase):
 	"""
 	Test a database self.driver for DB API 2.0 compatibility.
 	This implementation tests Gadfly, but the TestCase
-	is structured so that other self.drivers can subclass this 
-	test case to ensure compiliance with the DB-API. It is 
+	is structured so that other self.drivers can subclass this
+	test case to ensure compiliance with the DB-API. It is
 	expected that this TestCase may be expanded in the future
 	if ambiguities or edge conditions are discovered.
 
@@ -65,9 +65,9 @@ class test_dbapi20(unittest.TestCase):
 	self.driver, connect_args and connect_kw_args. Class specification
 	should be as follows:
 
-	import dbapi20 
+	import dbapi20
 	class mytest(dbapi20.DatabaseAPI20Test):
-		[...] 
+		[...]
 
 	__rcs_id__  = 'Id: dbapi20.py,v 1.10 2003/10/09 03:14:14 zenzen Exp'
 	__version__ = 'Revision: 1.10'
@@ -98,10 +98,10 @@ class test_dbapi20(unittest.TestCase):
 		try:
 			cur = con.cursor()
 			for ddl in (self.xddl1, self.xddl2):
-				try: 
+				try:
 					cur.execute(ddl)
 					con.commit()
-				except self.driver.Error: 
+				except self.driver.Error:
 					# Assume table didn't exist. Other tests will check if
 					# execute is busted.
 					pass
@@ -133,7 +133,7 @@ class test_dbapi20(unittest.TestCase):
 			# Must exist
 			threadsafety = self.driver.threadsafety
 			# Must be a valid value
-			self.failUnless(threadsafety in (0,1,2,3))
+			self.assertTrue(threadsafety in (0,1,2,3))
 		except AttributeError:
 			self.fail("Driver doesn't define threadsafety")
 
@@ -142,7 +142,7 @@ class test_dbapi20(unittest.TestCase):
 			# Must exist
 			paramstyle = self.driver.paramstyle
 			# Must be a valid value
-			self.failUnless(paramstyle in (
+			self.assertTrue(paramstyle in (
 				'qmark','numeric','named','format','pyformat'
 				))
 		except AttributeError:
@@ -151,13 +151,13 @@ class test_dbapi20(unittest.TestCase):
 	def test_Exceptions(self):
 		# Make sure required exceptions exist, and are in the
 		# defined heirarchy.
-		self.failUnless(issubclass(self.driver.InterfaceError,self.driver.Error))
-		self.failUnless(issubclass(self.driver.DatabaseError,self.driver.Error))
-		self.failUnless(issubclass(self.driver.OperationalError,self.driver.Error))
-		self.failUnless(issubclass(self.driver.IntegrityError,self.driver.Error))
-		self.failUnless(issubclass(self.driver.InternalError,self.driver.Error))
-		self.failUnless(issubclass(self.driver.ProgrammingError,self.driver.Error))
-		self.failUnless(issubclass(self.driver.NotSupportedError,self.driver.Error))
+		self.assertTrue(issubclass(self.driver.InterfaceError,self.driver.Error))
+		self.assertTrue(issubclass(self.driver.DatabaseError,self.driver.Error))
+		self.assertTrue(issubclass(self.driver.OperationalError,self.driver.Error))
+		self.assertTrue(issubclass(self.driver.IntegrityError,self.driver.Error))
+		self.assertTrue(issubclass(self.driver.InternalError,self.driver.Error))
+		self.assertTrue(issubclass(self.driver.ProgrammingError,self.driver.Error))
+		self.assertTrue(issubclass(self.driver.NotSupportedError,self.driver.Error))
 
 	def test_ExceptionsAsConnectionAttributes(self):
 		# OPTIONAL EXTENSION
@@ -169,15 +169,15 @@ class test_dbapi20(unittest.TestCase):
 		con = self._connect()
 		try:
 			drv = self.driver
-			self.failUnless(con.Warning is drv.Warning)
-			self.failUnless(con.Error is drv.Error)
-			self.failUnless(con.InterfaceError is drv.InterfaceError)
-			self.failUnless(con.DatabaseError is drv.DatabaseError)
-			self.failUnless(con.OperationalError is drv.OperationalError)
-			self.failUnless(con.IntegrityError is drv.IntegrityError)
-			self.failUnless(con.InternalError is drv.InternalError)
-			self.failUnless(con.ProgrammingError is drv.ProgrammingError)
-			self.failUnless(con.NotSupportedError is drv.NotSupportedError)
+			self.assertTrue(con.Warning is drv.Warning)
+			self.assertTrue(con.Error is drv.Error)
+			self.assertTrue(con.InterfaceError is drv.InterfaceError)
+			self.assertTrue(con.DatabaseError is drv.DatabaseError)
+			self.assertTrue(con.OperationalError is drv.OperationalError)
+			self.assertTrue(con.IntegrityError is drv.IntegrityError)
+			self.assertTrue(con.InternalError is drv.InternalError)
+			self.assertTrue(con.ProgrammingError is drv.ProgrammingError)
+			self.assertTrue(con.NotSupportedError is drv.NotSupportedError)
 		finally:
 			con.close()
 
@@ -273,21 +273,21 @@ class test_dbapi20(unittest.TestCase):
 			cur.execute("insert into %sbooze values ('Victoria Bitter')" % (
 				self.table_prefix
 			))
-			self.failUnlessEqual(cur.rowcount, 1,
+			self.assertEqual(cur.rowcount, 1,
 				'cursor.rowcount should == number or rows inserted, or '
 				'set to -1 after executing an insert statement'
 			)
 			cur.execute("insert into %sbooze select 'Victoria Bitter' WHERE FALSE" % (
 				self.table_prefix
 			))
-			self.failUnlessEqual(cur.rowcount, 0)
+			self.assertEqual(cur.rowcount, 0)
 			cur.execute("insert into %sbooze select 'First' UNION ALL select 'second'" % (
 				self.table_prefix
 			))
-			self.failUnlessEqual(cur.rowcount, 2)
+			self.assertEqual(cur.rowcount, 2)
 
 			cur.execute("select name from %sbooze" % self.table_prefix)
-			self.failUnlessEqual(cur.rowcount, -1,
+			self.assertEqual(cur.rowcount, -1,
 				'cursor.rowcount should == number of rows returned, or '
 				'set to -1 after executing a select statement'
 			)
@@ -376,7 +376,7 @@ class test_dbapi20(unittest.TestCase):
 		cur.execute("insert into %sbooze values ('Victoria Bitter')" % (
 			self.table_prefix
 		))
-		self.failUnless(cur.rowcount in (-1,1))
+		self.assertTrue(cur.rowcount in (-1,1))
 
 		if self.driver.paramstyle == 'qmark':
 			cur.execute(
@@ -405,7 +405,7 @@ class test_dbapi20(unittest.TestCase):
 			)
 		else:
 			self.fail('Invalid paramstyle')
-		self.failUnless(cur.rowcount in (-1,1))
+		self.assertTrue(cur.rowcount in (-1,1))
 
 		cur.execute('select name from %sbooze' % self.table_prefix)
 		res = cur.fetchall()
@@ -457,7 +457,7 @@ class test_dbapi20(unittest.TestCase):
 				)
 			else:
 				self.fail('Unknown paramstyle')
-			self.failUnless(cur.rowcount in (-1,2),
+			self.assertTrue(cur.rowcount in (-1,2),
 				'insert using cursor.executemany set cursor.rowcount to '
 				'incorrect value %r' % cur.rowcount
 			)
@@ -499,7 +499,7 @@ class test_dbapi20(unittest.TestCase):
 				'cursor.fetchone should return None if a query retrieves '
 				'no rows'
 			)
-			self.failUnless(cur.rowcount in (-1,0))
+			self.assertTrue(cur.rowcount in (-1,0))
 
 			# cursor.fetchone should raise an Error if called after
 			# executing a query that cannnot return rows
@@ -519,7 +519,7 @@ class test_dbapi20(unittest.TestCase):
 			self.assertEqual(cur.fetchone(),None,
 				'cursor.fetchone should return None if no more rows available'
 			)
-			self.failUnless(cur.rowcount in (-1,1))
+			self.assertTrue(cur.rowcount in (-1,1))
 		finally:
 			con.close()
 
@@ -575,7 +575,7 @@ class test_dbapi20(unittest.TestCase):
 				'cursor.fetchmany should return an empty sequence after '
 				'results are exhausted'
 			)
-			self.failUnless(cur.rowcount in (-1,6))
+			self.assertTrue(cur.rowcount in (-1,6))
 
 			# Same as above, using cursor.arraysize
 			cur.arraysize=4
@@ -588,12 +588,12 @@ class test_dbapi20(unittest.TestCase):
 			self.assertEqual(len(r),2)
 			r = cur.fetchmany() # Should be an empty sequence
 			self.assertEqual(len(r),0)
-			self.failUnless(cur.rowcount in (-1,6))
+			self.assertTrue(cur.rowcount in (-1,6))
 
 			cur.arraysize=6
 			cur.execute('select name from %sbooze' % self.table_prefix)
 			rows = cur.fetchmany() # Should get all rows
-			self.failUnless(cur.rowcount in (-1,6))
+			self.assertTrue(cur.rowcount in (-1,6))
 			self.assertEqual(len(rows),6)
 			self.assertEqual(len(rows),6)
 			rows = [r[0] for r in rows]
@@ -610,7 +610,7 @@ class test_dbapi20(unittest.TestCase):
 				'cursor.fetchmany should return an empty sequence if '
 				'called after the whole result set has been fetched'
 			)
-			self.failUnless(cur.rowcount in (-1,6))
+			self.assertTrue(cur.rowcount in (-1,6))
 
 			self.executeDDL2(cur)
 			cur.execute('select name from %sbarflys' % self.table_prefix)
@@ -619,7 +619,7 @@ class test_dbapi20(unittest.TestCase):
 				'cursor.fetchmany should return an empty sequence if '
 				'query retrieved no rows'
 			)
-			self.failUnless(cur.rowcount in (-1,0))
+			self.assertTrue(cur.rowcount in (-1,0))
 
 		finally:
 			con.close()
@@ -643,7 +643,7 @@ class test_dbapi20(unittest.TestCase):
 
 			cur.execute('select name from %sbooze' % self.table_prefix)
 			rows = cur.fetchall()
-			self.failUnless(cur.rowcount in (-1,len(self.samples)))
+			self.assertTrue(cur.rowcount in (-1,len(self.samples)))
 			self.assertEqual(len(rows),len(self.samples),
 				'cursor.fetchall did not retrieve all rows'
 			)
@@ -659,12 +659,12 @@ class test_dbapi20(unittest.TestCase):
 				'cursor.fetchall should return an empty list if called '
 				'after the whole result set has been fetched'
 			)
-			self.failUnless(cur.rowcount in (-1,len(self.samples)))
+			self.assertTrue(cur.rowcount in (-1,len(self.samples)))
 
 			self.executeDDL2(cur)
 			cur.execute('select name from %sbarflys' % self.table_prefix)
 			rows = cur.fetchall()
-			self.failUnless(cur.rowcount in (-1,0))
+			self.assertTrue(cur.rowcount in (-1,0))
 			self.assertEqual(len(rows),0,
 				'cursor.fetchall should return an empty list if '
 				'a select query returns no rows'
@@ -685,7 +685,7 @@ class test_dbapi20(unittest.TestCase):
 			rows23 = cur.fetchmany(2)
 			rows4  = cur.fetchone()
 			rows56 = cur.fetchall()
-			self.failUnless(cur.rowcount in (-1,6))
+			self.assertTrue(cur.rowcount in (-1,6))
 			self.assertEqual(len(rows23),2,
 				'fetchmany returned incorrect number of rows'
 			)
@@ -750,7 +750,7 @@ class test_dbapi20(unittest.TestCase):
 		con = self._connect()
 		try:
 			cur = con.cursor()
-			self.failUnless(hasattr(cur,'arraysize'),
+			self.assertTrue(hasattr(cur,'arraysize'),
 				'cursor.arraysize must be defined'
 				)
 		finally:
@@ -786,11 +786,11 @@ class test_dbapi20(unittest.TestCase):
 		try:
 			con.autocommit = True
 			# autocommit mode on, commit/abort on inappropriate.
-			self.failUnlessRaises(
+			self.assertRaises(
 				con.InterfaceError,
 				con.commit
 			)
-			self.failUnlessRaises(
+			self.assertRaises(
 				con.InterfaceError,
 				con.rollback
 			)
@@ -842,29 +842,40 @@ class test_dbapi20(unittest.TestCase):
 		b = self.driver.Binary(b'')
 
 	def test_STRING(self):
-		self.failUnless(hasattr(self.driver,'STRING'),
+		self.assertTrue(hasattr(self.driver,'STRING'),
 			'module.STRING must be defined'
 		)
 
 	def test_BINARY(self):
-		self.failUnless(hasattr(self.driver,'BINARY'),
+		self.assertTrue(hasattr(self.driver,'BINARY'),
 			'module.BINARY must be defined.'
 		)
 
 	def test_NUMBER(self):
-		self.failUnless(hasattr(self.driver,'NUMBER'),
+		self.assertTrue(hasattr(self.driver,'NUMBER'),
 			'module.NUMBER must be defined.'
 		)
 
 	def test_DATETIME(self):
-		self.failUnless(hasattr(self.driver,'DATETIME'),
+		self.assertTrue(hasattr(self.driver,'DATETIME'),
 			'module.DATETIME must be defined.'
 		)
 
 	def test_ROWID(self):
-		self.failUnless(hasattr(self.driver,'ROWID'),
+		self.assertTrue(hasattr(self.driver,'ROWID'),
 			'module.ROWID must be defined.'
 		)
 
+	def test_placeholder_escape(self):
+		con = self._connect()
+		try:
+			c = con.cursor()
+			c.execute("SELECT 100 %% %s", (99,))
+			self.assertEqual(1, c.fetchone()[0])
+			c.execute("SELECT 100 %% %(foo)s", {'foo': 99})
+			self.assertEqual(1, c.fetchone()[0])
+		finally:
+			con.close()
+
 if __name__ == '__main__':
 	unittest.main()
