@@ -84,13 +84,29 @@ sample_structured_parameters = [
 ]
 
 class test_iri(unittest.TestCase):
+	def testIP6Hosts(self):
+		"""
+		Validate that IPv6 hosts are properly extracted.
+		"""
+		s = [
+			('pq://[::1]/db', '::1'),
+			('pq://[::1]:1234/db', '::1'),
+			('pq://[1:2:3::1]/db', '1:2:3::1'),
+			('pq://[1:2:3::1]:1234/db', '1:2:3::1'),
+			('pq://[]:1234/db', ''),
+			('pq://[]/db', ''),
+		]
+		for i, h in s:
+			p = pg_iri.parse(i)
+			self.assertEqual(p['host'], h)
+
 	def testPresentPasswordObscure(self):
 		"password is present in IRI, and obscure it"
 		s = 'pq://user:pass@host:port/dbname'
 		o = 'pq://user:***@host:port/dbname'
 		p = pg_iri.parse(s)
 		ps = pg_iri.serialize(p, obscure_password = True)
-		self.failUnlessEqual(ps, o)
+		self.assertEqual(ps, o)
 
 	def testPresentPasswordObscure(self):
 		"password is *not* present in IRI, and do nothing"
@@ -98,11 +114,11 @@ class test_iri(unittest.TestCase):
 		o = 'pq://user@host:port/dbname'
 		p = pg_iri.parse(s)
 		ps = pg_iri.serialize(p, obscure_password = True)
-		self.failUnlessEqual(ps, o)
+		self.assertEqual(ps, o)
 
 	def testValueErrors(self):
 		for x in value_errors:
-			self.failUnlessRaises(ValueError,
+			self.assertRaises(ValueError,
 				pg_iri.parse, x
 			)
 
@@ -112,14 +128,14 @@ class test_iri(unittest.TestCase):
 			px = pg_iri.parse(x)
 			spx = pg_iri.serialize(px)
 			pspx = pg_iri.parse(spx)
-			self.failUnless(
+			self.assertTrue(
 				pspx == px,
 				"parse-serialize incongruity, %r -> %r -> %r : %r != %r" %(
 					x, px, spx, pspx, px
 				)
 			)
 			spspx = pg_iri.serialize(pspx)
-			self.failUnless(
+			self.assertTrue(
 				spx == spspx,
 				"parse-serialize incongruity, %r -> %r -> %r -> %r : %r != %r" %(
 					x, px, spx, pspx, spspx, spx
@@ -130,7 +146,7 @@ class test_iri(unittest.TestCase):
 		for x in sample_structured_parameters:
 			xs = pg_iri.serialize(x)
 			uxs = pg_iri.parse(xs)
-			self.failUnless(
+			self.assertTrue(
 				x == uxs,
 				"serialize-parse incongruity, %r -> %r -> %r" %(
 					x, xs, uxs,
@@ -138,7 +154,4 @@ class test_iri(unittest.TestCase):
 			)
 
 if __name__ == '__main__':
-	from types import ModuleType
-	this = ModuleType("this")
-	this.__dict__.update(globals())
-	unittest.main(this)
+	unittest.main()
